@@ -17,7 +17,6 @@
  */
 
 #include "kerncompat.h"
-#include "androidcompat.h"
 
 #include <stdio.h>
 #include <stdio_ext.h>
@@ -31,11 +30,11 @@
 
 #include "kernel-lib/list.h"
 #include "kernel-lib/radix-tree.h"
-#include "ctree.h"
-#include "extent-cache.h"
-#include "disk-io.h"
-#include "volumes.h"
-#include "transaction.h"
+#include "kernel-shared/ctree.h"
+#include "common/extent-cache.h"
+#include "kernel-shared/disk-io.h"
+#include "kernel-shared/volumes.h"
+#include "kernel-shared/transaction.h"
 #include "crypto/crc32c.h"
 #include "common/utils.h"
 #include "check/common.h"
@@ -194,8 +193,7 @@ static struct btrfs_chunk *create_chunk_item(struct chunk_record *record)
 	return ret;
 }
 
-static void init_recover_control(struct recover_control *rc, int verbose,
-		int yes)
+static void init_recover_control(struct recover_control *rc, int yes)
 {
 	memset(rc, 0, sizeof(struct recover_control));
 	cache_tree_init(&rc->chunk);
@@ -208,7 +206,7 @@ static void init_recover_control(struct recover_control *rc, int verbose,
 	INIT_LIST_HEAD(&rc->rebuild_chunks);
 	INIT_LIST_HEAD(&rc->unrepaired_chunks);
 
-	rc->verbose = verbose;
+	rc->verbose = bconf.verbose;
 	rc->yes = yes;
 	pthread_mutex_init(&rc->rc_lock, NULL);
 }
@@ -635,12 +633,12 @@ bg_check:
 	l = path.nodes[0];
 	slot = path.slots[0];
 	bg_ptr = btrfs_item_ptr(l, slot, struct btrfs_block_group_item);
-	if (chunk->type_flags != btrfs_disk_block_group_flags(l, bg_ptr)) {
+	if (chunk->type_flags != btrfs_block_group_flags(l, bg_ptr)) {
 		if (rc->verbose)
 			fprintf(stderr,
 				"Chunk[%llu, %llu]'s type(%llu) is different with Block Group's type(%llu)\n",
 				chunk->offset, chunk->length, chunk->type_flags,
-				btrfs_disk_block_group_flags(l, bg_ptr));
+				btrfs_block_group_flags(l, bg_ptr));
 		btrfs_release_path(&path);
 		return -ENOENT;
 	}
@@ -1087,7 +1085,7 @@ err:
 static int block_group_free_all_extent(struct btrfs_trans_handle *trans,
 				       struct block_group_record *bg)
 {
-	struct btrfs_block_group_cache *cache;
+	struct btrfs_block_group *cache;
 	struct btrfs_fs_info *info;
 	u64 start;
 	u64 end;
@@ -1097,8 +1095,8 @@ static int block_group_free_all_extent(struct btrfs_trans_handle *trans,
 	if (!cache)
 		return -ENOENT;
 
-	start = cache->key.objectid;
-	end = start + cache->key.offset - 1;
+	start = cache->start;
+	end = start + cache->length - 1;
 
 	if (list_empty(&cache->dirty_list))
 		list_add_tail(&cache->dirty_list, &trans->dirty_bgs);
@@ -1358,9 +1356,9 @@ static int __insert_block_group(struct btrfs_trans_handle *trans,
 	struct btrfs_key key;
 	int ret = 0;
 
-	btrfs_set_block_group_used(&bg_item, used);
-	btrfs_set_block_group_chunk_objectid(&bg_item, used);
-	btrfs_set_block_group_flags(&bg_item, chunk_rec->type_flags);
+	btrfs_set_stack_block_group_used(&bg_item, used);
+	btrfs_set_stack_block_group_chunk_objectid(&bg_item, used);
+	btrfs_set_stack_block_group_flags(&bg_item, chunk_rec->type_flags);
 	key.objectid = chunk_rec->offset;
 	key.type = BTRFS_BLOCK_GROUP_ITEM_KEY;
 	key.offset = chunk_rec->length;
@@ -1447,7 +1445,7 @@ open_ctree_with_broken_chunk(struct recover_control *rc)
 	fs_info->is_chunk_recover = 1;
 
 	fs_info->fs_devices = rc->fs_devices;
-	ret = btrfs_open_devices(fs_info->fs_devices, O_RDWR);
+	ret = btrfs_open_devices(fs_info, fs_info->fs_devices, O_RDWR);
 	if (ret)
 		goto out;
 
@@ -1574,10 +1572,7 @@ static int btrfs_get_device_extents(u64 chunk_object,
 
 static int calc_num_stripes(u64 type)
 {
-	if (type & (BTRFS_BLOCK_GROUP_RAID0 |
-		    BTRFS_BLOCK_GROUP_RAID10 |
-		    BTRFS_BLOCK_GROUP_RAID5 |
-		    BTRFS_BLOCK_GROUP_RAID6))
+	if (btrfs_bg_type_is_stripey(type))
 		return 0;
 	else if (type & (BTRFS_BLOCK_GROUP_RAID1 |
 			 BTRFS_BLOCK_GROUP_DUP))
@@ -1590,14 +1585,6 @@ static int calc_num_stripes(u64 type)
 		return 1;
 }
 
-static inline int calc_sub_nstripes(u64 type)
-{
-	if (type & BTRFS_BLOCK_GROUP_RAID10)
-		return 2;
-	else
-		return 1;
-}
-
 static int btrfs_verify_device_extents(struct block_group_record *bg,
 				       struct list_head *devexts, int ndevexts)
 {
@@ -1660,13 +1647,11 @@ static int btrfs_calc_stripe_index(struct chunk_record *chunk, u64 logical)
 	} else if (chunk->type_flags & BTRFS_BLOCK_GROUP_RAID10) {
 		index = stripe_nr % (chunk->num_stripes / chunk->sub_stripes);
 		index *= chunk->sub_stripes;
-	} else if (chunk->type_flags & BTRFS_BLOCK_GROUP_RAID5) {
-		nr_data_stripes = chunk->num_stripes - 1;
-		index = stripe_nr % nr_data_stripes;
-		stripe_nr /= nr_data_stripes;
-		index = (index + stripe_nr) % chunk->num_stripes;
-	} else if (chunk->type_flags & BTRFS_BLOCK_GROUP_RAID6) {
-		nr_data_stripes = chunk->num_stripes - 2;
+	} else if (chunk->type_flags & BTRFS_BLOCK_GROUP_RAID56_MASK) {
+		int parity;
+
+		parity = btrfs_bg_type_to_nparity(chunk->type_flags);
+		nr_data_stripes = chunk->num_stripes - parity;
 		index = stripe_nr % nr_data_stripes;
 		stripe_nr /= nr_data_stripes;
 		index = (index + stripe_nr) % chunk->num_stripes;
@@ -1760,8 +1745,7 @@ no_extent_record:
 	if (list_empty(&devexts))
 		return 0;
 
-	if (chunk->type_flags & (BTRFS_BLOCK_GROUP_RAID5 |
-				 BTRFS_BLOCK_GROUP_RAID6)) {
+	if (chunk->type_flags & BTRFS_BLOCK_GROUP_RAID56_MASK) {
 		/* Fixme: try to recover the order by the parity block. */
 		list_splice_tail(&devexts, &chunk->dextents);
 		return -EINVAL;
@@ -1795,11 +1779,6 @@ no_extent_record:
 	return 0;
 }
 
-#define BTRFS_ORDERED_RAID	(BTRFS_BLOCK_GROUP_RAID0 |	\
-				 BTRFS_BLOCK_GROUP_RAID10 |	\
-				 BTRFS_BLOCK_GROUP_RAID5 |	\
-				 BTRFS_BLOCK_GROUP_RAID6)
-
 static int btrfs_rebuild_chunk_stripes(struct recover_control *rc,
 				       struct chunk_record *chunk)
 {
@@ -1811,10 +1790,10 @@ static int btrfs_rebuild_chunk_stripes(struct recover_control *rc,
 	 * is we can reorder the stripes in the system metadata chunk.
 	 */
 	if ((chunk->type_flags & BTRFS_BLOCK_GROUP_METADATA) &&
-	    (chunk->type_flags & BTRFS_ORDERED_RAID))
+	    btrfs_bg_type_is_stripey(chunk->type_flags))
 		ret =btrfs_rebuild_ordered_meta_chunk_stripes(rc, chunk);
 	else if ((chunk->type_flags & BTRFS_BLOCK_GROUP_DATA) &&
-		 (chunk->type_flags & BTRFS_ORDERED_RAID))
+		 btrfs_bg_type_is_stripey(chunk->type_flags))
 		ret = 1;	/* Be handled after the fs is opened. */
 	else
 		ret = btrfs_rebuild_unordered_chunk_stripes(rc, chunk);
@@ -1877,11 +1856,7 @@ static u64 calc_data_offset(struct btrfs_key *key,
 	data_offset = key->offset + csum_offset * blocksize - chunk->offset;
 	nr_data_stripes = chunk->num_stripes;
 
-	if (chunk->type_flags & BTRFS_BLOCK_GROUP_RAID5)
-		nr_data_stripes -= 1;
-	else if (chunk->type_flags & BTRFS_BLOCK_GROUP_RAID6)
-		nr_data_stripes -= 2;
-
+	nr_data_stripes -= btrfs_bg_type_to_nparity(chunk->type_flags);
 	logical_stripe_nr = data_offset / chunk->stripe_len;
 	dev_stripe_nr = logical_stripe_nr / nr_data_stripes;
 
@@ -1912,7 +1887,7 @@ static int check_one_csum(int fd, u64 start, u32 len, u32 tree_csum,
 	}
 	ret = 0;
 	put_unaligned_le32(tree_csum, expected_csum);
-	btrfs_csum_data(csum_type, (u8 *)data, result, len);
+	btrfs_csum_data(NULL, csum_type, (u8 *)data, result, len);
 	if (memcmp(result, expected_csum, csum_size) != 0)
 		ret = 1;
 out:
@@ -2193,8 +2168,8 @@ static int btrfs_rebuild_ordered_data_chunk_stripes(struct recover_control *rc,
 	u8 flags;
 
 	list_for_each_entry_safe(chunk, next, &rc->unrepaired_chunks, list) {
-		if ((chunk->type_flags & BTRFS_BLOCK_GROUP_DATA)
-		 && (chunk->type_flags & BTRFS_ORDERED_RAID)) {
+		if ((chunk->type_flags & BTRFS_BLOCK_GROUP_DATA) &&
+		    btrfs_bg_type_is_stripey(chunk->type_flags)) {
 			flags = 0;
 			err = rebuild_raid_data_chunk_stripes(rc, root, chunk,
 							      &flags);
@@ -2246,7 +2221,7 @@ static int btrfs_recover_chunks(struct recover_control *rc)
 		chunk->io_width = BTRFS_STRIPE_LEN;
 		chunk->io_align = BTRFS_STRIPE_LEN;
 		chunk->sector_size = rc->sectorsize;
-		chunk->sub_stripes = calc_sub_nstripes(bg->flags);
+		chunk->sub_stripes = btrfs_bg_type_to_sub_stripes(bg->flags);
 
 		ret = insert_cache_extent(&rc->chunk, &chunk->cache);
 		if (ret == -EEXIST) {
@@ -2319,14 +2294,14 @@ static void validate_rebuild_chunks(struct recover_control *rc)
 /*
  * Return 0 when successful, < 0 on error and > 0 if aborted by user
  */
-int btrfs_recover_chunk_tree(const char *path, int verbose, int yes)
+int btrfs_recover_chunk_tree(const char *path, int yes)
 {
 	int ret = 0;
 	struct btrfs_root *root = NULL;
 	struct btrfs_trans_handle *trans;
 	struct recover_control rc;
 
-	init_recover_control(&rc, verbose, yes);
+	init_recover_control(&rc, yes);
 
 	ret = recover_prepare(&rc, path);
 	if (ret) {
